\
\ Code for the DCC command station.
\
\ $ProjectVersion: 0.297 $
\
\ This program is in charge of:
\    * generating a valid DCC signal from commands given on the serial line
\    * do some reporting on the serial line
\
\ References:
\    * NMRA DCC Standards and Recommended Practices:
\      http://www.tttrains.com/nmradcc/

include picisr.fs
include piceeprom.fs
include libnibble.fs
include picflash.fs
include libstrings.fs

\ ----------------------------------------------------------------------
\ Pinout used in this program
\ ----------------------------------------------------------------------

0 pin-b dcc-output
1 pin-b led-empty
2 pin-b led-error

\ ----------------------------------------------------------------------
\ Serial port handling
\ ----------------------------------------------------------------------

:: emit ( b -- ) >w begin txif bit-set? until w> txreg ! ;

: space ( -- ) $20 emit ;
: crlf ( -- ) $d emit $a emit ;
: print ( -- ) begin str-char dup while emit repeat drop ;

\ The following word could be written as `print crlf' but this would add
\ one more return-stack depth level.
: print-crlf ( -- ) begin str-char dup while emit repeat drop crlf ;

variable verbose
macro
: verbose? verbose @ 1 and ;
target

macro
: v-print-crlf ( -- ) verbose? if print-crlf then ;
: v-space ( -- ) verbose? if space then ;
target

macro
: clear-errors ( -- ) oerr bit-set? if cren bit-clr cren bit-set then ;
: key? ( -- f ) rcif bit-set? ;
target

: key ( -- b )
  begin key? until
  rcreg @ clear-errors
  verbose? 0= if exit then
  dup emit
;

: get8 ( -- b ) key hex>nibble 4 lshift key hex>nibble or ;
: put8 ( b -- ) dup 4 rshift nibble>hex emit $f and nibble>hex emit ;

\ ----------------------------------------------------------------------
\ Interrupt driven DCC pulse generator
\ ----------------------------------------------------------------------

\ set-timer will wait for 3.2*(256-n)-0.2 s for a 20MHz oscillator and a
\ 16 prescaler. For ones, use n=238 (57.4s => 1.03% error). For zero, use
\ values between n=0 (819.0s) and n=224 (102.2s).

macro
: set-timer ( n -- ) tmr0 +! ;
target

\ delay contains the duration of the next low pulse when the high pulse
\ is in progress or 0 otherwise

variable delay

\ pulse and half-pulse work the reverse way around: they set the timer
\ to the proper value, then call the idle-loop to do other tasks, then
\ wait for the timer to expire and immediately change the DCC output
\ state.  This way, it is permitted to take some time (up to 40s,
\ i.e. around 200 instructions) to handle secondary tasks. More
\ complex tasks can be handled if the zero flag is true (in which case
\ we have up to 80s, i.e. around 400 instructions).

macro
: half-pulse ( -- ) delay @ set-timer ;          \ Set timer to pulse value
target

:: pulse ( duration -- ) delay ! half-pulse clrwdt ;

: bit-zero ( -- ) $64 pulse ;
: bit-one ( -- )  $ee pulse ;

\ Data being transmitted. We implement a 16 bits shift register, left-aligned,
\ so that preamble, start and stop bits can be easily implemented. To write
\ something into the DCC generator, wait for data-bits to be zero, write
\ into data-high and data-low (left aligned) and set data-bits to the correct
\ number of bits.

variable data-high
variable data-low
variable data-bits

\ Internal data. Data is double buffered, to let the time to the main routine
\ to write the needed bytes. Idling takes place automatically if the main
\ routine fails to write a proper bit sequence into data-high, data-low and
\ data-bits.

variable dcc-high
variable dcc-low
variable dcc-bits

variable idling?

: idling-step ( -- )
  1 idling? +!
  idling? @ 1 = if $ff dcc-high ! $fc dcc-low ! $f dcc-bits ! exit then \ Pre
  idling? @ 2 = if $ff dcc-high ! 0 dcc-low ! 9 dcc-bits ! exit then \ All 1
  idling? @ 3 = if $0 dcc-high ! 0 dcc-low ! 9 dcc-bits ! exit then  \ All 0
  0 idling? ! $ff dcc-high ! $80 dcc-low ! 9 dcc-bits !              \ Checksum
;

: start-idling ( -- ) led-empty high 0 idling? ! idling-step ;

: dcc-refill ( -- )
  idling? @ if idling-step exit then
  data-bits @ 0= if start-idling exit then
  data-high @ dcc-high ! data-low @ dcc-low ! data-bits @ dcc-bits !
  0 data-bits ! led-empty low
;

\ isr-tmr0 is called when there is a timer overflow interrupt
\ This word is very long to avoid return stack depth exhaustion, but very
\ thoroughly commented. It terminates with a call to isr-exit to properly
\ exit from interrupt

isr : isr-tmr0 ( -- )
    t0if bit-clr                        \ Acknowledge interrupt
    delay @ if
	dcc-output low                  \ Drive DCC output low
	half-pulse                      \ Rearm timer
	0 delay !                       \ Note low pulse in progress
    else
      dcc-output high                     \ Drive DCC output high
      dcc-low rlf! dcc-high rlf!          \ Shift data left through carry
      c bit-set? if bit-one else bit-zero then   \ Send bit and clear watchdog
      1 dcc-bits -!                       \ Decrement data length
      z bit-set? if dcc-refill then       \ Refill transmit buffer
    then
    isr-restore-return
;

\ ----------------------------------------------------------------------
\ DCC encoder configuration
\ ----------------------------------------------------------------------

eevariable cfg-repeat
eecreate cfg-magic 2 eeallot

: explain-var ( n -- ) dup put8 [char] : space ee@ put8 space print crlf ;

: configuration-help ( -- )
  cfg-repeat c" Number of times to repeat a command" explain-var
;

: factory-defaults ( -- )
  3 cfg-repeat ee!
  $27 cfg-magic ee!
  $05 cfg-magic 1+ ee!
;

: check-magic ( -- )
  cfg-magic ee@ $27 = if cfg-magic 1+ ee@ $05 = if exit then then
  c" Bad configuration data, reverting to factory defaults" v-print-crlf
  factory-defaults
;

\ ----------------------------------------------------------------------
\ DCC data manipulation
\ ----------------------------------------------------------------------

variable checksum

\ Wait for the transmission slot to be available

macro
: wait-for-slot ( -- ) begin data-bits @ 0= until ;
target

\ Transmit a preamble (14 bits to 1) and a start bit (one bit to 0)

: preamble ( -- )
    wait-for-slot 0 checksum ! $ff data-high ! $fc data-low ! $f data-bits ! ;

\ send-byte sends a byte followed by a start bit. send-final-byte ends
\ the packet with a stop bit after the byte. send-checksum sends the checksum
\ with a stop bit.

: finish-byte ( b -- ) dup checksum xor! data-high ! 9 data-bits ! ;
:: send-byte ( b -- ) wait-for-slot $00 data-low ! finish-byte ;
: send-final-byte ( b -- ) wait-for-slot $80 data-low or! finish-byte ;
: send-checksum ( -- ) checksum @ send-final-byte ;

\ General reset command

: dcc-reset ( -- ) preamble 0 send-byte 0 send-byte send-checksum ;

\ Send a packet located at cmd with length in cmd-length

variable cmd-length
create cmd 5 allot

variable cmd-tmp
: dcc-packet ( -- )
  preamble cmd-length @ cmd-tmp !
  cmd @ send-byte 1 cmd-tmp -! z bit-set? if send-checksum exit then
  cmd 1+ @ send-byte 1 cmd-tmp -! z bit-set? if send-checksum exit then
  cmd 2 + @ send-byte 1 cmd-tmp -! z bit-set? if send-checksum exit then
  cmd 3 + @ send-byte 1 cmd-tmp -! z bit-set? if send-checksum exit then
  cmd 4 + @ send-byte send-checksum
;

variable repeat-tmp
: dcc-packet-repeat ( -- ) cfg-repeat ee@ repeat-tmp v-for dcc-packet v-next ;

\ ----------------------------------------------------------------------
\ Commands handling
\ ----------------------------------------------------------------------

: ack ( -- ) v-space [char] ! emit v-space ;

: get8/w ( -- b/w ) v-space get8 >w ;
macro
: get8w get8/w w> ;
target

: get-dcc-packet ( -- )
  get8w cmd ! 1 cmd-tmp -! z bit-set? if exit then
  get8w cmd 1+ ! 1 cmd-tmp -! z bit-set? if exit then
  get8w cmd 2 + ! 1 cmd-tmp -! z bit-set? if exit then
  get8w cmd 3 + ! 1 cmd-tmp -! z bit-set? if exit then
  get8w cmd 4 + !
;

: cmd-any-packet ( -- )
  get8 dup cmd-length ! cmd-tmp !
  get-dcc-packet
  ack dcc-packet-repeat
;

: cmd-emergency-stop ( -- )
  2 cmd-length ! get8 cmd ! $41 cmd 1+ ! ack dcc-packet-repeat
;

: >speed-n ( b -- s ) 1- c bit-clr rrf-tos c bit-set? if $10 or then $42 + ;
: >speed ( b -- s ) dup if >speed-n exit then drop $40 ;

: cmd-fwd ( -- )
  2 cmd-length ! get8 cmd ! v-space get8 >speed $20 or cmd 1+ !
  ack dcc-packet-repeat
;

: cmd-bwd ( -- )
  2 cmd-length ! get8 cmd ! v-space get8 >speed cmd 1+ !
  ack dcc-packet-repeat
;

: cmd-setcfg ( -- ) get8 v-space get8 ack swap ee! ;

: cmd-getcfg ( -- ) get8 ack ee@ put8 ;

: cmd-reset ( -- ) ack dcc-reset ;

: cmd-verbon ( -- ) 1 verbose ! ack ;
: cmd-verboff ( -- ) 0 verbose ! ack ;

: cmd-config-help ( -- ) ack configuration-help ;

\ $Format: ": version c\" DCC generator $ProjectVersion$\" ;"$
: version c" DCC generator 0.297" ;

: cmd-help ( -- )
  ack crlf version print-crlf
  c" f loco speed            Move loco in forward direction" print-crlf
  c" b loco speed            Move loco in backward direction" print-crlf
  c" s loco                  Emergency stop" print-crlf
  c" r                       Emergency reset" print-crlf
  c" c nn                    Get configuration parameter" print-crlf
  c" C nn vv                 Set configuration parameter" print-crlf
  c" v                       Turn verbose mode on" print-crlf
  c" V                       Turn verbose mode off" print-crlf
  c" a length xx yy zz ...   Send any valid DCC packet" print-crlf
  c" h                       List of configuration parameters" print-crlf
  c" F                       Factory defaults" print-crlf
  c" ?                       Help (this message)" print-crlf
;

: cmd-factory-reset ( -- ) ack factory-defaults ;

: nack ( c -- c )
  verbose? if
    crlf c" ??? Unknown-command: `" print dup emit [char] ' emit exit
  then
  [char] ? emit
;

: handle-command ( c -- c )
  dup [char] f = if cmd-fwd exit then
  dup [char] b = if cmd-bwd exit then
  dup [char] s = if cmd-emergency-stop exit then
  dup [char] r = if cmd-reset exit then
  dup [char] c = if cmd-getcfg exit then
  dup [char] C = if cmd-setcfg exit then
  dup [char] v = if cmd-verbon exit then
  dup [char] V = if cmd-verboff exit then
  dup [char] a = if cmd-any-packet exit then
  dup [char] h = if cmd-config-help exit then
  dup [char] F = if cmd-factory-reset exit then
  dup [char] ? = if cmd-help exit then
  nack
;

\ ----------------------------------------------------------------------
\ Initialization
\ ----------------------------------------------------------------------

: init ( -- )
    \ Data
    $ff dcc-high  !              \ 16 bits to one will help the system
    $ff dcc-low !                \ to enter permanent mode
    $10 dcc-bits !
    \ Port B
    $00 portb !                  \ Clear output latches
    $f8 trisb !                  \ Ports 0, 1 and 2 as output
    \ Serial port transmit (Microchip DS30292C page 100)
    portc 6 >output              \ Port C6 is TX
    \ $14 spbrg !                  \ 57600 bauds with BGRH high at 20MhZ
    $19 spbrg !                  \ 9600 bauds with BGRH high at 4MhZ
    $24 txsta !                  \ Asynchronous transmit high-speed generator
    \ Serial port receive (Microchip DS30292C page 102)
    $90 rcsta !                  \ Serial-enable 8bits-rx continuous-receive
    1 verbose !                  \ Enable verbose mode
    \ Timer
    $03 option_reg !             \ CLKOUT, prescaler to timer 0, prescaler=16
    t0ie bit-set                 \ Enable timer 0 overflow interrupt
    $f0 tmr0 !                   \ Do not wait too long before first interrupt
    enable-interrupts
;

: greetings ( -- ) version v-print-crlf ;

\ ----------------------------------------------------------------------
\ Main program and main loop
\ ----------------------------------------------------------------------

: prompt ( -- )
  verbose? if crlf c" Command (type `?' for help) > " print exit then
  [char] > emit ;

: mainloop ( -- )
  begin
    prompt key handle-command drop
  again ;

main : main ( -- ) init check-magic greetings dcc-reset mainloop ;

\ ----------------------------------------------------------------------
\ Configuration word
\ ----------------------------------------------------------------------

fosc-hs set-fosc    \ High-speed oscillator
false set-/pwrte    \ Power-up reset timer
false set-lvp       \ No need for low voltage programming
